var log4js = require( "log4js" ),
    crypto = require( "crypto" ),
    _log = log4js.getLogger( "models.collection.archive" ),
    _performance = require( "../lib/utils/performance.js" ),
    _items = [];

var self = module.exports = {

    /**
     * Intialize the collection with the given items.
     * @param {Object} items
     */
    init: function( items ) {
        _items = items;
    },

    /**
     * Set the given item at the given id.
     *
     * Overrides the item at the id if already exists.
     *
     * @param {String} id
     * @param {Object} data
     */
    setItem: function( id, data ) {
        _items[ id ] = data;
    },

    /**
     * Get all items of the collection.
     * @returns {Object}
     */
    getItems: function( ) {
        return _items;
    },

    /**
     * Returns a special result for rendering the list in the frontend.
     *
     * If a valid cache key if given the result contains no items.
     *
     * @param   {Integer} page     page number to show
     * @param   {Integer} cacheKey cache key if available
     * @param   {Integer} numberOfItems optional - number of items per page, default=50
     * @param   {object} selection optional - selection criteria - user_id, event_id, final, recent, only_framename
     * @returns {Object}  the result as a object
     */
    getListItems: function( page, cacheKey, numberOfItems, selection ) {
    	// selection
//	socket.on("get_archive", function(user_id, event_id, final, recent, only_framename) {

        _log.debug( "called get archive action" );

        var keys = Object.keys( _items ).reverse(),
            count = keys.length,
            items_selected = [];

        var startItems = _performance();
		var ts_limit = Date.now() - 1000 * 60 * 60 * 24 * 14; // within two weeks from now
        for ( var n = 0; n < count; n++ ) {
            var item = _items[ keys[ n ] ];
            if (selection && selection.recent && item.ts < ts_limit) continue
 			if (selection && selection.event_id >= 0 && (!str.event_data || str.event_data.event_id != event_id)) continue;
           
            items_selected.push( {
                _id: item._id,
                name: item.name,
                ts: item.ts,
                face_id: item.face_id,
                frame_id: item.frame_id,
                n_shots: item.n_shots
            } );
        }
        _log.debug( "finished selecting items (" + _performance( startItems ) + "ms)" );

        var keys = Object.keys( items_selected ),
            count = keys.length,
            itemsPerPage = numberOfItems || 50,
            items = [],
            hashString = "",
            pages = Math.ceil( count / itemsPerPage );

        if ( page !== undefined ) {
            page = parseInt( page, 10 );
        } else {
            page = 1;
        }
        if ( page > pages ) {
            page = 1;
        }
        if ( page < 1 ) {
            page = pages;
        }
        _log.debug( "start selecting items for page number " + page );
        var startItems = _performance();
        for ( var n = ( page - 1 ) * itemsPerPage; ( n < page * itemsPerPage ) && ( n < count ); n++ ) {
            var key_selected = keys [ n ],
            	item = items_selected[ key_selected ];
            items.push( item );
            hashString += key_selected;
        }
        var result = {
            count: count,
            page: page,
            pageCount: pages,
            itemsPerPage : itemsPerPage,
            data: undefined,
            cacheKey: crypto.createHash( "md5" ).update( hashString ).digest( "hex" ),
            cacheValid: false
        };
        _log.debug( "finished selecting page items (" + _performance( startItems ) + "ms)" );

        if ( cacheKey === undefined || cacheKey !== result.cacheKey ) {

            // return result with data...
            result.data = items;
        } else {
            result.cacheValid = true;
        }
        return result;
    },

    /**
     * Returns the number of items in the collection.
     * @returns {Integer} number of items
     */
    count: function( ) {
        return Object.keys( _items ).length;
    },

    renameString: function( id, name ) {
        if ( _items[ id ] !== undefined ) {
            _items[ id ].name = name;
            return true;
        }
        return false;
    },

    deleteString: function( id ) {
        if ( _items[ id ] !== undefined ) {
            delete _items[ id ];
            return true;
        }
        return false;
    }
};
