/*******************************************************************************
 viewer_maintarget.js: client module for main target display

 $Author: dmc $
 $Revision: 2510 $
 $Date: 2019-03-04 17:08:52 -0400 (Mon, 04 Mar 2019) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smtv2-lodi/public/javascripts/viewer_maintarget.js $
*******************************************************************************/

// handles drawing main target on cur_string()

/* hold-off reticle:

hold_active: user has control, draw focused crosshair
 - true on 'press'
 - false on 'pressup' and 'panend'

args.pan_hold: client side crosshair position
 - is ahead of server when user is moving it
 - equals hold_active
 - when true, pan_hold_x and pan_hold_y are the crosshair position

string.shot_params.hold: server side crosshair position
 - true on 'panend'
 - false on 'tap'
 - hold_x, hold_y
     when true, the server's crosshair position
     when true and args.pan_hold is false, the client's crosshair position
     
hold movements updates necessary:
 - maintarget.update_canvas (for crosshair active status and position)
 - plotting.update_canvas (only for crosshair position)

press/pan event sequence:
hold and release: press, pressup
hold, pan, release: press, panstart, pan ..., panend
*/

var maintarget = new function() {

	this.init = function() {
		this.div = $("#divMainTarget"); // html container
	}
	
	this.clear = function() {
		hide(this.div);
		this.div.empty();
	}

	this.canvas = null; // maintain canvas for efficiency
	this.events = null;
	
	// for panning
	this.pan_x = 0;
	this.pan_y = 0;
	this.set_pan = function(x, y) {
		if (x != x || y != y) return; // do not pan to NaN
		this.pan_x = x;
		this.pan_y = y;
	}
	
	// for zooming
	this.zoom = this.zoom_default = 7; // moa default
	this.set_zoom = function(z) {
		// min zoom level = 20 times real bullet size
		// factor should match bullet scale limit in client_face
		var face = cur_string().face();
		var bullet_diameter = 7.82;

		this.zoom_min = 20 * bullet_diameter * face.mm_to_moa;
		
		// max zoom level = 4 times targetboard width
		this.zoom_max = face.data.targetboard.width * face.mm_to_moa * 4;

		var zoom_last = this.zoom;

		if (z < this.zoom_min) z = this.zoom_min;
		else if (z > this.zoom_max) z = this.zoom_max;
		else if (!(z >= this.zoom_min && z <= this.zoom_max)) z = this.zoom_default;
		
		this.zoom = z;
		this.args.update_px_scale();
		
		// adjust pan so that center of screen is preserved
		var zoom_delta = (this.zoom / zoom_last);
		this.set_pan(this.pan_x / zoom_delta, this.pan_y / zoom_delta);
	}

	// for passing parameters to face/hammer
	// minimum initialization required for set_pan
	this.args = {
		w: 0,
		h: 0
	};
	this.args.update_px_scale = function() {
		this.moa_to_px = Math.min(this.w, this.h) / maintarget.zoom;
		this.mm_to_px = this.face.mm_to_moa * this.moa_to_px;
	}
	
	this.hold_active = false; // hold-off reticle is currently active (after a long press)
	
	this.drag_calibration_active = false; // shot calibration by dragging is active (after a long press)

	// update main target
	this.update = function() {
		this.clear();
		if (!cur_string()) return;

		var args = this.args;
		args.string = cur_string();
		args.frame = args.string.frame();
		args.face = args.string.face();
		
		//log("maintarget.update(" + args.frame.data.name + ", " + args.face.data.name + ", (" + this.div.width() + "x" + this.div.height() + "))");

		// error check because this is hard to track down
		if (this.div.height() == 0) {
			alert("MAINTARGET DIV HEIGHT IS ZERO!");
			return;
		}

		// set size of display
		args.w = this.div.width() - 5;
		args.h = this.div.height() - 5;
		
		if (printing) {
			args.w = 600;
			args.h = 600;
		}
		
		// check if existing canvas needs to be resized
		if (ok(this.canvas)) {
			if (this.canvas[0].width != args.w || this.canvas[0].height != args.h) {
				this.canvas[0].width = args.w;
				this.canvas[0].height = args.h;
				this.set_pan(0, 0);
			}
		}

		// rebuild canvas if it does not already exist
		if (!ok(this.canvas)) {
			
			//log("rebuilding maintarget canvas");
			
			this.canvas = $("<canvas>").attr({width: args.w, height: args.h});
			this.events = new Hammer(this.canvas[0]);
			
			this.events.on("tap", function(e) {
				var args = maintarget.args;
				
				// in drag calibration mode, select closest shot to tap
				if (toolbar_system.visualcal_mode) {
					
					var x = (e.center.x - e.target.offsetLeft - args.cx) / args.mm_to_px;
					var y = -(e.center.y - e.target.offsetTop - args.cy) / args.mm_to_px;
					
					var shots = cur_string().shots();
					
					var min_distsq = 1000000; // within 1 meter
					var found = -1;
					
					for (var i in shots) {
						var shot = shots[i];
						if (!shot || !shot.valid) continue; // valid, on target only
						var distsq = (shot.x - x) * (shot.x - x) + (shot.y - y) * (shot.y - y);
						if (distsq < min_distsq) {
							found = i;
							min_distsq = distsq;
						}
					}
					
					if (found != -1) {
						cur_string().select_shot(found);
						update();
					}
				
				// simulate a shot
				} else if (toolbar_system.simulate_mode) {
					if (!args.frame.device()) alert(S("ADMIN_SIMULATE_UNASSOCIATED"));
					else {
						var x = (e.center.x - e.target.offsetLeft - args.cx) / args.mm_to_px;
						var y = -(e.center.y - e.target.offsetTop - args.cy) / args.mm_to_px;
						log("simulating shot at " + x + "," + y);
						client_simulate_shot_xy(x,y); // in client.js
					}
					
				// tap disabled params hold
				} else if (plotting.enable_controls && args.string.data.shot_params.hold) {
					var params = args.string.data.shot_params;
					if (params.hold) {
						params.hold = false;
						args.string.crud({
							shot_params: {
								hold: false,
							}
						});
						maintarget.update_canvas();
						plotting.update();
					}
					
				// center main target on cursor
				} else {

					// center view at touch location
					//maintarget.set_pan(e.center.x - args.w/2, e.center.y - args.h/2);
					//maintarget.update_canvas_now();
					//plotting.update_canvas();
					
					// zoom in
					maintarget.set_zoom(maintarget.zoom / 1.25);
					update();
				}
					
			});
			
			// for panning (touch or mouse)
			this.events.get("pan").set({ direction: Hammer.DIRECTION_ALL });
			this.events.on("panstart", function(e) {
				var args = maintarget.args;
				
				args.pan_last_x = 0;
				args.pan_last_y = 0;
				
				// hold crosshair state is initialized by 'press'
				
			});
			this.events.on("pan", function(e) {
				var args = maintarget.args;

				var dx = e.deltaX - args.pan_last_x; args.pan_last_x = e.deltaX;
				var dy = e.deltaY - args.pan_last_y; args.pan_last_y = e.deltaY;
				
				// move display of the dragged shot
				if (args.drag_cal_active) {
					args.drag_cal_x += dx * 0.5 / args.mm_to_px;
					args.drag_cal_y -= dy * 0.5 / args.mm_to_px;
				
				// adjust hold-off reticle clientside
				} else if (maintarget.hold_active) {
					args.pan_hold_x += dx * 0.5 / args.moa_to_px;
					args.pan_hold_y -= dy * 0.5 / args.moa_to_px;

				// pan main target
				} else {
					maintarget.set_pan(maintarget.pan_x + dx, maintarget.pan_y + dy);
				}
				
				maintarget.update_canvas();
				plotting.update_canvas();
			});
			this.events.on("panend", function(e) {
				var args = maintarget.args;
				
				// finish dragging shot
				if (args.drag_cal_active) {
					
					var shot = cur_string().data.shots[args.drag_cal_shot];
					if (shot) {
						
						var mx = shot.x + args.drag_cal_x;
						var my = shot.y + args.drag_cal_y;
						
						shot.measured = true;
						shot.measured_x = mx;
						shot.measured_y = my;
						
						cur_string().crud_shot(args.drag_cal_shot, {
							measured: true, measured_x: mx, measured_y: my
						});
	
						update_calibration(); // client side calculation
					}
					
					args.drag_cal_active = false;

					maintarget.update_canvas_now();

				// finish moving hold crosshair
				} else if (maintarget.hold_active) {
					maintarget.hold_active = false;
						
					// apply local setting to params
					var params = args.string.data.shot_params;
					params.hold_x = args.pan_hold_x;
					params.hold_y = args.pan_hold_y;
					
					// send to server
					args.string.crud({
						shot_params: {
							hold: true,
							hold_x: params.hold_x,
							hold_y: params.hold_y
						}
					});
					
					args.pan_hold = false;
					
					maintarget.update_canvas_now();
					plotting.update_canvas();
				}
			});

			// for pinch zoom on tablet
			this.events.get("pinch").set({enable: true});
			this.events.on("pinchstart", function(e) {
				maintarget.args.zoom_pinchstart = maintarget.zoom;
			});
			this.events.on("pinch", function(e) {
				var args = maintarget.args;
				
				maintarget.set_zoom(maintarget.args.zoom_pinchstart / e.scale);
				
				maintarget.update_canvas();
				plotting.update_canvas();
			});
			this.events.on("pinchend", function(e) {
				update(); // full update because target row uses zoom
			});
	
			// for enabling hold-off reticle
			this.events.on("press", function(e) {
				var args = maintarget.args;
				
				//alert(maintarget.zoom + " " + maintarget.pan_x + " " + maintarget.pan_y);
				
				// drag calibration when in admin frame mode
				if (toolbar_system.visualcal_mode) {
					
					// start dragging the current shot (local display only)
					if (cur_string().cur_shot()) {
						args.drag_cal_active = true;
						args.drag_cal_shot = cur_string().cur_shot_index;
						args.drag_cal_x = 0;
						args.drag_cal_y = 0;
					}
					
					maintarget.update_canvas_now();

/* disabling hold crosshair for now, because it may conflict with multiple shooters
				} else if (plotting.enable_controls) {
					maintarget.hold_active = true;
					
					var params = args.string.data.shot_params;

					// no hold set - reset
					if (!params.hold) {
						params.hold = true;
						params.hold_x = 0;
						params.hold_y = 0;
					}

					args.pan_hold = true; // for local display
					args.pan_hold_x = params.hold_x;
					args.pan_hold_y = params.hold_y;
					
					maintarget.update_canvas_now();
*/
				} else {
				
					// zoom out
					maintarget.set_zoom(maintarget.zoom * 1.25);
					update();

				}
			});
			this.events.on("pressup", function(e) { // press cancelled without pan
				var args = maintarget.args;

				// disable drag cal without applying changes
				args.drag_cal_active = false;
			
				if (maintarget.hold_active) {
					maintarget.hold_active = false;
					args.pan_hold = false;
					
					// sync the current hold (in case it was just created)
					var params = args.string.data.shot_params;
					args.string.crud({
						shot_params: {
							hold: true,
							hold_x: params.hold_x,
							hold_y: params.hold_y
						}
					});

				}
				
				maintarget.update_canvas_now();
			});

			// for zooming with scroll wheel on PC
			this.div.off("wheel");
			this.div.on("wheel", function(e) {
				if (e.originalEvent.deltaY > 0) {
					maintarget.set_zoom(maintarget.zoom * 1.15);
				} else {
					maintarget.set_zoom(maintarget.zoom / 1.15);
				}
				update(); // full update because target row uses zoom
			});

			this.set_zoom(this.zoom);
		}

		// add canvas to page
		this.div.append(this.canvas);
		
		this.update_canvas_now();

		//log("maintarget.update() done");
		show(this.div);
	};
	

	// use requestAnimationFrame if it is supported
	// otherwise use the clunky timer implementation
	this.update_canvas = function() {
		if (window.requestAnimationFrame) {
			window.requestAnimationFrame(function() {
				maintarget.update_canvas_now();
			});
		} else {
			this.update_canvas_with_timer();
		}
	}

	// fallback timer implementation for old devices
	this.timer = null;
	this.update_needed = false;
	this.update_canvas_with_timer = function() {
		if (this.timer == null) {
			this.update_canvas_now();
			this.update_needed = false;
			this.timer = setTimeout(function() {
				maintarget.timer = null;
				if (maintarget.update_needed) {
					maintarget.update_canvas_now();
					maintarget.update_needed = false;
				}
			}, 250);
		} else {
			this.update_needed = true;
		}
	}

	this.update_canvas_now = function() {
		if (!ok(this.canvas)) return;
		if (!ok(this.canvas[0].parentElement)) return;

		var g = this.canvas[0].getContext("2d");
		var args = this.args;

		// update pan/zoom
		args.cx = args.w/2 + this.pan_x;
		args.cy = args.h/2 + this.pan_y;
		args.update_px_scale();

		g.clearRect(0, 0, args.w, args.h);
		
		var fclass = draw_fclass_target()
		
		// draw layers in order
		args.face.draw_target_paper(g, args, fclass);
		
		args.face.draw_target_grid(g, args); // will decide whether or not to draw grid, and what size
		args.face.draw_stats_circle(g, args);
		args.face.draw_stats_center(g, args);
		// draw shots last, so they are on top:
		args.face.draw_target_shots(g, args, fclass);

		var params = args.string.data.shot_params;

		// draw hold reticle (plotting)
		if (live_mode() && (maintarget.hold_active || args.pan_hold || params.hold)) {

			// draw cross at hold position
			g.beginPath();

			var px = args.cx;
			var py = args.cy;
			
			if (args.pan_hold) {
				px += args.moa_to_px * args.pan_hold_x;
				py -= args.moa_to_px * args.pan_hold_y;
			} else if (params.hold) {
				px += args.moa_to_px * params.hold_x;
				py -= args.moa_to_px * params.hold_y;
			}

			var cross = args.moa_to_px * ((maintarget.hold_active)?(2):(1));
			g.moveTo(px, py - cross); g.lineTo(px, py + cross);
			g.moveTo(px - cross, py); g.lineTo(px + cross, py);

			g.strokeStyle = "#8F8";
			g.lineWidth = args.moa_to_px * ((maintarget.hold_active)?(0.06):(0.03));
			g.stroke();

		}
		
	}


}


