/*******************************************************************************
 beam.js: utilities for talking to a V1.1 and V1.2 sensor beam

 Usage:
  - create a BeamsListener (one for whole system; handles mult. targets)
  - each unique IP address (i.e. yellow box) gets a "TargetListener".
  - a Targetlistener:
    -- polls its IP address (and someday soon, applies some "smarts" to this too!)
    -- is an EventEmitter: 'tgt_heartbeat', 'meas-from-beam'
    that IP address and collects shot data.
  - more than one "TargetFrame" may be associated with each TargetListener (that is up
    to the caller).

  - call .get_listener(targetframe) for each targetframe - it returns a TargetListener
  - connect event handlers to this TargetListener
  - if desired, turn each of them on/off using .start_poller() / .stop_poller()
  - process events it produces:
    'tgt_heartbeat' - emitted with member tgt_stat: 'tgt_there'every time a response 
        is received from the target.  Emitted with member tgt_stat:  'tgt_missedpacket'
        each time a packed is missed (turn on "yellow"?)
    'meas-from-beam' - emitted for every shot; includes shot#, coords, timestamp.


**** previous****
 Usage:
  - create a BeamsListener (one for whole system; handles mult. targets)
  - call .get_listener(ip) for each target.
  - connect event handlers to these
  - if desired, turn each of them on using .start_poller() / .stop_poller()
  - process events it produces:
    'tgt_heartbeat' - emitted with member tgt_stat: 'tgt_there'every time a response 
        is received from the target.  Emitted with member tgt_stat:  'tgt_missedpacket'
        each time a packed is missed (turn on "yellow"?)
    'meas-from-beam' - emitted for every shot; includes shot#, coords, timestamp.

 
 $Author: dmc $
 $Revision: 3221 $
 $Date: 2023-11-16 18:06:33 -0400 (Thu, 16 Nov 2023) $
 $HeadURL: http://support2.silvermountaintargets.com/svn/smt/smt.common.lib/beam.js $
*******************************************************************************/


var logdebugjs = require('./logdebug.js');
var logdebug = logdebugjs.logdebug; // the logdebug() method
var logerror = logdebugjs.logerror;
//var logdebug = console.log;

var log = require('log4js').getLogger('beam.js');

log.setLevel('TRACE');
//log.setLevel('OFF');

var util=require('util')
var events = require("events");
var dg = require('dgram') // node.js datagram (i.e. UDP)
var fs = require('fs');
// 
//require('buffertools'); // adds new methods to class 'Buffer'

if (0) { // enable this is using Launchpad to generate hardware timing signals

// open /dev/ttyACM0, if available
var launchpad_wr_strm = fs.createWriteStream('/dev/ttyACM0',
    {'flags': 'w', encoding: 'ascii'/*, bufferSize: 1*/});
//console.log("fs.createWriteStream /dev/ttyACM0 =",launchpad_wr_strm);
//console.log('fs.createWriteStream ,launchpad_wr_strm.writable=',launchpad_wr_strm.writable);
//console.log('fs.createWriteStream ,launchpad_wr_strm.readable=',launchpad_wr_strm.readable);
var launchpad_rd_strm = fs.createReadStream('/dev/ttyACM0',
    {'flags': 'r+', encoding: 'ascii', bufferSize: 10});
//
// "r+" -> lsof:
//COMMAND   PID     USER   FD   TYPE DEVICE SIZE/OFF     NODE NAME
//node    11632      500    8u   CHR  166,0      0t0 40865220 /dev/ttyACM0
//
//console.log("fs.createReadStream /dev/ttyACM0 =",launchpad_rd_strm);
//console.log('fs.createReadStream ,launchpad_rd_strm.writable=',launchpad_rd_strm.writable);
//console.log('fs.createReadStream ,launchpad_rd_strm.readable=',launchpad_rd_strm.readable);
launchpad_rd_strm.on('data', function(d1) {
  console.log('Launchpad: ',d1);
});

//var launchpad_fd;
//fs.open("/dev/ttyACM0","r+",function (e,fd) {
//  if (!e) {
//    launchpad_fd = fd;
//  }
//console.log("fs.open /dev/ttyACM0 e=",e,"fd=",fd);
//});


} // end 'if (0)'


function BeamsListener() {
    if(false === (this instanceof BeamsListener)){
        return new BeamsListener();
    }
    events.EventEmitter.call(this);
}

util.inherits(BeamsListener, events.EventEmitter);


exports.beamlisteners = {};  // a hash, indexed by IP, used by server.

exports.BeamsListener = BeamsListener;

var IP_PORT_UDP_SHOT_REQUEST=10000		//! Port zur Schussanfrage an der Messchiene
var IP_PORT_TCP_SHOT_REQUEST=11000		//! #TE:2015-01-11 port for tcp communication
// google xlate: Port for shot request to the measurement Rails

//var IP_PORT_UDP_TARGET_REQUEST=10001		// IP table stuff
var IP_PORT_UDP_ADMIN_REQUEST=10002		//! Admin-Port der Messchiene
var IP_PORT_UDP_PANEL_RECV=9999		//! Antwort-Port am Panel
var UDP_MSG_LENGTH=128; // beam might always axpect 128-byt chunks?

//commands:
var HI_THERE=0xDE;	    //!< Hallo Messchiene!
var SHOT_REQUEST=0xDD; // "get new shot"; sent to IP_PORT_UDP_SHOT_REQUEST
// google xlate:  request as if new shot

//#define SHOT_NEXT_TIME		    0x05	    //!< Wiederholung Schuss letzte Nr.
var SHOT_NEXT_TIME=0x05	    // get shot # (int16)
// google xlate: repeat last shot No.
var ERR_SHOT_NR=0x11;  // response: (google xlate) repeat shot -> No not Available

//#define EMPTY_IP_TABLE          0x0A        //!< Komplette IP Table leeren.
//#define REMOVE_IP               0x0B        //!< Bestimmte IP entfernen.
//#define SET_IP_ADDRESS          0xC1        //!< Neue IP Adresse setzen.
//#define SET_MAC_ADDRESS         0xC2        //!< Neue MAC Adresse setzen.
//#define SET_EEPROM_DATA			0xC3		//!< Kompletten EEPROM setzen.
//#define GET_IP_ADDRESS          0xD1        //!< IP Adresse holen.

// DMC: GET_MAC_ADDRESS does not seem to be impletemented by the beam:
//var GET_MAC_ADDRESS=0xD2;       //!< MAC Adresse holen.
var GET_EEPROM_DATA=0xD3;		//!< Kompletten EEPROM holen.



// responses:
var ANSWER_OK=0x1A;	    //!< OK
var NO_SHOT=0xFF;	    //!< Antwort keiner da
var SHOT_L=0xAF;	    // 24 Apr 2013 - new firmware!!
var SHOT_R=0xEF;	    // 24 Apr 2013 - new firmware!!
// #TE 2014-02-01:
// SHOT_L is returned when we have a new shot and the Running Boar is moving to the left.
// SHOT_R is returned when we have a new shot and the Running Boar is moving to the right.
//
// 2013-6-8 firmware change:
//
// - firstly, the "3+ readings" logic has been removed.  A measurement will be reported
// if at least one mic fires.
//
// - secondly, to support quick (and hopefully transparent!) recovery from situations
//   in which a packet is lost and the server does not receive the "SHOT" message from
//   the beam, the "NO_SHOT" response has been enhanced.  It used to return "0xffff" if
//   as a response when 
//


// response: new shot detected.  Payload contains shot data.

var SHOT=0x01;	    //!< Antwort Schuss da

// the low five bits on "out_set" record whether a valid reading was
// obtained, or if the counter on that channel overflower.
//
// bitmasks for "out_set":
var MIC1 = 0x01;
var MIC2 = 0x02;
var MIC3 = 0x04;
var MIC4 = 0x08;
var MIC5 = 0x10;
var ALL_MICS = MIC1 | MIC2 | MIC3 | MIC4 | MIC5;
var MIC_MASKS = [MIC1, MIC2, MIC3, MIC4, MIC5];

// listener for all incoming udp packets, from all beams
function my_l1(msg/*Buffer*/, rinfo) {
  // crash-protection:
  if (typeof msg != 'object' || typeof rinfo != 'object' || typeof msg.length != 'number' || msg.length < 2) {
    return;
  }

  var len = msg.length-2;
  var stat = stat1 = msg[0];
  var stat2 = msg[1]; // second status byte
  var res;
  var ip = rinfo.address;

  // find the beam (by IP address) that this request came from
  var my_beam = exports.beamlisteners[ip];
//console.log("dmc my_l1 msg",msg,"ip",ip,"my_beam",my_beam,"qq",exports.beamlisteners);
  if (!my_beam) { // uh-oh...
    return; // silently ignore...
  }

  my_beam.tickle(msg,rinfo);

  logdebug(8,"my_l1 stat1=",stat,'stat2=',stat2,"+",len,
      "bytes=",msg.toString("hex",2)+", from "+rinfo.address+":"+rinfo.port);

  if (stat == ANSWER_OK) {
    res = msg.toString('utf8',2); // should contain beam name
    logdebug("****ANSWER_OK**** ",len,"res=",res)
  } else if (stat == NO_SHOT || stat == SHOT_L || stat == SHOT_R || stat == SHOT) {
    my_beam.handle_shot(msg); // decodes the shot, detect missing shots, note boar direction, etc.
  } else {
    logdebug("**my_l1 Unknown Status",stat,stat2,len,
      "bytes=",msg.toString("hex",2)+", from "+rinfo.address+":"+rinfo.port);
  }

  if (0 && stat!="NO_SHOT") {
    logdebug("my_l1 NO_SHOT stat=",stat,'stat2=',stat2,"+",len,
      "bytes=",msg.toString("hex",2)+", from "+rinfo.address+":"+rinfo.port);
  }
  if (0 && stat=="NO_SHOT") {
    logdebug("my_l1 NO_SHOT stat=",stat,'stat1=',stat1,'stat2=',stat2,'('+msg.toString("hex",0,2)+')',"+",len,
      "bytes=",msg.toString("hex",2)+", from "+rinfo.address+":"+rinfo.port);
  }

};


var udpsock;
//
//
exports.debug = 0



function init_socket() {
  if (typeof udpsock !== 'undefined') {
    return;
  }
  udpsock = dg.createSocket('udp4',my_l1);
  udpsock.on("error", function (exx) {
    logerror('UDPsock: error', exx);
  });
  // socket does not show up in "netstat -an" until bind() is done
  udpsock.bind(IP_PORT_UDP_PANEL_RECV); // responses back to this port
}


// DMC 28 Feb 2013: even though we are now recording shot data to a database,
// let's continue to independently log all shot data to a plain ASCII log file.
//
//var fs=require('fs');
//var wr_stream_shotslog =
//   fs.createWriteStream('shotslog.txt', {'flags': 'a', encoding: 'ascii'});
//console.log('fs.createWriteStream  ,wr_stream_shotslog.writeable=',wr_stream_shotslog.writeable);

BeamsListener.prototype.delete_listener = function( device ) {
  if ( exports.beamlisteners[ device.internalId ] !== undefined ) {
    exports.beamlisteners[ device.internalId ] = undefined;
  }
}

// get (or add) a listener for a targetframe
// 
BeamsListener.prototype.get_listener = function( device ) {
//console.log("BeamsListener.get_listener",device.internalId);
  if ( 
    device.connectionType === undefined 
    && ( device.type === "G2" || device.smtdevtype === "G2" || device.type === "SOLO"  || device.type === "GSDP" ) 
  ) {
    device.connectionType = "TCP";
  } else if ( device.connectionType === undefined ) {
    device.connectionType = "UDP";
  }

    if (device.internalId === undefined) { 	// bugfix 23-Oct-2018: this gets done for all device types
        device.internalId = device.ip;
//console.log("BeamsListener.get_listener  device.internalId === undefined; set to ",device.ip);
//        console.log("  device.internalId === undefined; set to ",device.ip);
    }


	if ( device.connectionType !== "TCP" ) {
    init_socket();
//    console.log("we are a G1,",device.internalId);
  }

  var d = exports.beamlisteners[ device.internalId ]; // maybe it's already defined....?
  if (d == undefined) {
  	//log("creating new listener for " + ip);
    d = new TargetListener( device.ip, device.connectionType, device.port );
    exports.beamlisteners[ device.internalId ] = d;
//console.log("BeamsListener.get_listener created new ",d,"for id",device.internalId);
  } else {
  	//console.log("found listener for " + ip);
  }
  return d;
}

// stop (completely shutdown) listener.
function stop_listener(tlistener) {
  var beamip = tlistener.ip;
  logdebug(4,"%%%%%%%%%% stop_listener",beamip,exports.beamlisteners[beamip]);
    
  tlistener.stop_poller(); // clear interval timer, delete tcpClient
  tlistener.beamevents.removeAllListeners();
}

exports.stop_all_listeners = function() {

  Object.keys(exports.beamlisteners).forEach(function (beamip, idx, arr) {
    stop_listener(exports.beamlisteners[beamip]);
    // don't delete... keep it in its idling state
    // delete exports.beamlisteners[beamip];
// DMC TODO: delete beamlisteners if TCP?
  });

  if (typeof udpsock !== 'undefined') {
    udpsock.close();
    udpsock = undefined;
  }
}





// for internal use only:
function send_cmd(cmd,tlis,beam_port,extra_bytes) {
  if (!udpsock) {
    return;  // can't do anything
  }
   var xtrabytes = 0;
   if (extra_bytes !== undefined) {
      xtrabytes=extra_bytes.length;
   }
   var nbytes_w=2+xtrabytes;
   var my_sendbuf = new Buffer(nbytes_w); my_sendbuf.fill(0,2);
   my_sendbuf.fill(cmd,0,2); // first two bytes get opcode
   if (xtrabytes != 0) {
      extra_bytes.copy(my_sendbuf,2);
   }
   var beam_ip = tlis.ip;
   tlis.send_time = Date.now();
//console.log("writing command="+my_sendbuf.toString('hex',0,1)+" to beam "+beam_ip+":"+beam_port+"\n");
   if (exports.debug) {
//      console.log("writing command="+my_sendbuf.toString('hex',0,1)+" to beam "+beam_ip+":"+beam_port+"\n");
   }
   var dropsome = 0;
//   dropsome = 1;
   if(dropsome && Math.random() < 0.1) {
     logdebug("dropping packet************");
   } else {
     udpsock.send(my_sendbuf,0,nbytes_w,beam_port,beam_ip);
   }
}




//// will send event "answer_ok" if there is a response
//exports.ping = function(beamip) {
//  send_cmd(HI_THERE,beamip,IP_PORT_UDP_ADMIN_REQUEST);
//}


//send_cmd(GET_EEPROM_DATA,beamip,IP_PORT_UDP_ADMIN_REQUEST);


//send_cmd(SHOT_REQUEST,beamip,IP_PORT_UDP_SHOT_REQUEST);

//TargetListener = {};
//TargetListener.prototype.TargetListener = function() {

var objnum = 1;

/**
 * #TE:2015-01-11 add connection type ('UDP' or 'TCP'), if given type is undefined, use UDP connection by default
 */
function TargetListener( beamip, type, port ) {
  if (! (this instanceof TargetListener) ) return new TargetListener();
  this.ip = beamip;
  this.recv_time = 0;
  this.port = port;
  if (exports.debug) {
    logdebug("DEBUG: TargetListener ctor, ip=",this.ip);
  }
  this.oldshots = []; // array of (decoded) shots received from beam, indexed by beam's shot# (0..250)
  this.beam_readings = []; // array of shot messages (27-byte Buffer) received from beam, indexed by beam's shot# (0..250)
  this.shotnums_wanted = []; // zero or nonzero; indexed by shotnum
  this.getshot_n_trycounter = []; // indexed by shotnum; 
  this.beamevents = new events.EventEmitter();
  this.tcpClient = undefined;

  // #TE:2015-01-11 add connection type ('UDP' or 'TCP')
  this.connectionType = 'UDP';
  if (type == 'TCP') {
    this.connectionType = 'TCP';
  }
// DMC: I think this belongs in start_poller, stop_poller
if (0) {
  if (type !== undefined && type === 'TCP') {
    this.initTcpClient(beamip);
  }
}
  
  events.EventEmitter.call(this);
}
util.inherits(TargetListener, events.EventEmitter);
exports.TargetListener = TargetListener;

/**
 * #TE:2015-01-11
 * Initialize a tcp client and connect to the given beam ip.
 */
TargetListener.prototype.initTcpClient = function(beamip) {
  log.info('initTcpClient #',++g_initTcpClient_cnt,beamip);
  var self = this;
  var Client = require('./lib/connection/tcp').Client;
  var port = IP_PORT_TCP_SHOT_REQUEST;
  
  if (this.port !== undefined) {
    port = this.port;
  }

  this.connectionType = 'TCP';
  this.tcpClient = new Client({
    port: port, 
    host: beamip,
    tcpClient_num: g_initTcpClient_cnt,
    connectedStateOnConnect: ( port === IP_PORT_TCP_SHOT_REQUEST )
  });
  this.lastShotNumber = undefined;

  var handleIncomingShot = function(data) {
    if (self.lastShotNumber === undefined || self.lastShotNumber === data.shotNumber - 1) {
      self.lastShotNumber = data.shotNumber;
    }
    log.info('self.lastshotnumber: ' + self.lastShotNumber + ' shot.shotnumber: ' + data.shotNumber);
      
    var shot = {
      shotnum:data.shotNumber, 
      counters:[
        data.counter0,
        data.counter1,
        data.counter2,
        data.counter3,
        data.counter4,
        data.counter5,
        data.counter6,
        data.counter7,
      ],
      raw_counters: data.raw_counters,
      G2_temp_rdg:data.temperature, // shot.temperature for G2; could be undefined
      out_set: 0,
      GPIO18: data.GPIO18, // JRPL target position reading
      mcucount: data.shotCountVector
    };

    if (data.mic_mapping) {
      shot.mic_mapping = data.mic_mapping;
    }
    //console.log("DMC beam.js trace 2, ",shot.G2_temp_rdg, n );
    self.evt_time = Date.now();
    self.elapsed_time = self.evt_time - self.send_time
    self.evt = "shot";

    var logshot = {
      ts: self.evt_time,
      //ip: self.ip, //targetid: this.targetid,
      source: 'G2',
      shot: shot, //geom: this.geom
      //connection_type: 'TCP',
    }
    self.emit_shot(logshot); // G2: initTcpClient
  };
    
  this.tcpClient.on('new-shot', function(data) { 
//      console.log('new-shot');
      handleIncomingShot(data); 
  });
  this.tcpClient.on('get-shot', function(data) { 
      if (data.result) {
          handleIncomingShot(data.shot);
      }
      if (data.number > self.lastShotNumber) {
          self.lastShotNumber = data.number;
      }
//      console.log('get-shot self.lastshotnumber: ' + self.lastShotNumber + ' shot.shotnumber: ' + data.shotNumber);
  });
    
/*
  this.tcpClient.on('connected', function() {
    console.log('connected');
  });
    
  this.tcpClient.on('disconnected', function() {
    console.log('disconnected');
  });
*/
    
  this.tcpClient.on('ping-request', function(data) {
    //console.log('ping-request(' + data + ')');
    if (self.lastShotNumber !== undefined && self.lastShotNumber < data) {
        for (n = self.lastShotNumber + 1; n <= data; n++) {
            self.tcpClient.send('get-shot', { number: n });
        }
    }
  });
  
  log.info("now starting tcp client");

  setInterval(function() {
    // tcpClient.connected indicates whether a connection is deemed to be open
    // tcpClient.awaitPingCounter is the number of outstanding pings
//    if (self.tcpClient !== undefined && self.tcpClient.connected)
    if (self.tcpClient !== undefined && self.tcpClient.connected && self.tcpClient.awaitPingCounter < 6)
      self.tgt_stat = 'tgt_there';
    else
      self.tgt_stat = 'tgt_missedpacket';
    self.beamevents.emit('tgt_heartbeat');
  }, 1500);
  this.tcpClient.connect();
}

// binary data for one shot is in Buffer "msg".  Create a
// Javascript object with this decoded info
exports.decode_shot = TargetListener.prototype.decode_shot = function decode_shot(msg/*Buffer*/) {
/* 25 bytes:
typedef struct {
	uint16_t	number; // DMC: shot number 0..249, then wrap to 0
	uint32_t	counter1;
	uint32_t	counter2;
	uint32_t	counter3;
	uint32_t	counter4;
	uint32_t	counter5;
	uint16_t	temperature_counts; // DMC: temperature A/D counts
	uint8_t		out_set;
} shot_t;
*/

  // if we don't have enough bytes in the message, one of the "read" methods
  // below will throw an exception.
  var shot = undefined;
  if (msg.length < 27) {
    return shot; // return 'undefined'
  }
  var x=2; // skip first two bytes (message status)
  var number=msg.readUInt16LE(x); x+=2;
  var counters=[
    msg.readUInt32LE(x)
    ,msg.readUInt32LE(x+=4)
    ,msg.readUInt32LE(x+=4)
    ,msg.readUInt32LE(x+=4)
    ,msg.readUInt32LE(x+=4)
  ]; x+=4; // 5 counters
  var temperature_counts=msg.readUInt16LE(x); x+=2;
  var out_set=msg.readUInt8(x); x+=1;
  var shot = {
    shotnum:number, counters:counters,
    temp_rdg:temperature_counts, out_set: out_set
  };
  return shot;
}

TargetListener.prototype.save_shot = function(logshot) { // write 'logshot' to logfile
  //wr_stream_shotslog.write(JSON.stringify(logshot)+"\n")
}

TargetListener.prototype.emit_shot = function(logshot) { // emit 'logshot' to listeners
  this.beamevents.emit('meas-from-beam',logshot);
}

// if a missed shot is detected, handle_missed_shots() will
// queue up a list of shots that need to be fetched
TargetListener.prototype.handle_missed_shots = function(stat2) {
  // an obscure edge case: if we lose a shot, but the beam 'wraps' and goes past 0 before we can get it.
  if (this.expected_shotnum > stat2 && stat2 != 0) {
    this.expected_shotnum = 0; // best we can do; we've shots [expected_shotnum..249] are irretrievable
  }

  // OK, we have missed [expected_shotnum..stat2-1]
  var jj; var jjmax = (stat2 == 0) ? 250 : stat2;
  for (jj = this.expected_shotnum; jj < jjmax; jj++) this.shotnums_wanted[jj] = 1;
  logdebug('handle_missed_shots - refetch',this.expected_shotnum,'..',jjmax);
}

TargetListener.prototype.clear_missed_shots = function() {
  var jj;
  for (jj = 0; jj < 250; jj++) this.shotnums_wanted[jj] = 0;
  logdebug('clear_missed_shots');
}


var null_msg_buffer_23 = new Buffer(23);
null_msg_buffer_23.fill(0);

TargetListener.prototype.handle_shot = function(msg/*Buffer*/) { // when called, 'this' points to
  // the TargetListener object that belongs (by IP address) to the incoming message.
  // incoming message might be one of the "NO_SHOT" responses, or it might be "SHOT".
  // It is possible that we've missed a shot.  This is where we handle that case.
  //
  // design decisions:
  // - with new beam firmware, we'll do things nicely.
  // - with old beam firmware, we do not yet implement logic to detect missed shots.  Later,
  //   we might be able to do it; possibly we will detect a missed shot only when another later
  //   shot is fired.  Might be able to revisit this later.
  //



  //
  // the get_shot_n operation (opcode SHOT_NEXT_TIME=0x05) has several behaviours, all of which
  // we should be able to handle here.
  // in a freshly-powered-up beam that has made no measurements, the response code is
  // 0x01 ('SHOT'), followed by
  //   0x00 (2013-6-8 firmware: 'next shot will be 00') or 0x01 (pre 2013-6-8 firmware: response code repeated),
  //   plus 25 bytes of 0x00 (a null reading).


  // after at least one shot has been measured, the response to a get_shot_n(6) where shot#6
  // has not yet been created is:
  // (new 2013-6-8 firmware):
  // 0xff ('NO_SHOT') 0x02 ('next shot will be 02'), plus a couple of garbage bytes.
  // (pre 2013-6-8 firmware):
  // 0xff ('NO_SHOT') 0xff (response code repeated), plus a couple of garbage bytes.

  // measured shotnums run from 0-249, inclusive.
  //
  // After shot #249 has been measured, the shot counter wraps to 0.  After this,
  // doing a get_shot_n(j) where j > current shot num, results in a "NO_SHOT" message
  // being returned - i.e., old shots are now inaccessible.  This might be a problem.


  var stat1 = msg[0]; // first status byte.  Will be "SHOT", or one of the "NO_SHOT*" variants
  var stat2 = msg[1]; // second status byte

  logdebug(6, 'handle_shot(',this.ip,') stat==>',stat1,stat2);

  if (stat1 == SHOT_L) {
    this.direction = "left";
    stat1 = SHOT;  // #TE 2014-02-01: we have a new shot with direction left, so set stat1 to SHOT
  } else if (stat1 == SHOT_R) {
    this.direction = "right";
    stat1 = SHOT;  // #TE 2014-02-01: we have a new shot with direction right, so set stat1 to SHOT
  }
  
  if (stat1 != SHOT) { // then we have one of the "NO_SHOT" messages
    // in the old, pre-2013-6-8 firmware, 'stat2' will be 0xFF
    if (stat2 == NO_SHOT) { // old firmware - we can't do any intelligent missed-shot detection
      return;
    }
    this.new_firmware = true;

    // new firmware: have we missed a shot?
    if (this.expected_shotnum == undefined) return;
    if (this.expected_shotnum == stat2) return; // we have not missed a shot

    this.handle_missed_shots(stat2); // generate list of shots to fetch
    return; // we're finished with all of the "NO_SHOT*" cases
  }

  // From here on, stat1 == SHOT

  // see if beam has not measured any shots since it was powered up
  // (we need to see if, except for the first four bytes, all bytes are zero


  // workaround the need for buffertools:
  function is_23bytes_0(bb) { // Late-night, klugey-as-hell... but, somehow, it works!!
    var i;
    for (i = 0; i < 23; i++) {
      if (bb[i] != 0) return false;
    }
    return true;
  }


//  if (msg.length == 27 && null_msg_buffer_23.xxequals(msg.slice(4))) { // tested; this logic works
  if (msg.length == 27 && is_23bytes_0(msg.slice(4))) {
    this.clear_missed_shots();
    this.expected_shotnum = 0;
    return; // ignore shot, it's not real
  }

  // each shot has a 'shotnum'.  Sometimes the shotnum is expected:
  //  - we've done a get_shot_n(j), and shot #j arrives
  //  - we've done a get_new_shot(), shot #j arrives, and shot #j-1 had been the most recent shot
  //
  // in other cases, the shotnum is unexpected'.
  //  - we've done a get_new_shot(), shot #j arrives, but the previous most recent shotnum != j-1.  Perhaps
  //    a packet was lost/missed.  Perhaps power to the beam has been cycled.

  var shot = this.decode_shot(msg);

  // shot will be 'undefined' if msg was too short (<27bytes)
  if (!shot) {
    return;
  }

  // TE 2014-01-26: we need the last direction info from the beam for the shot...
  shot.direction = this.direction;
  
  if (this.new_firmware && this.expected_shotnum != undefined
  && this.expected_shotnum != shot.shotnum) { // missed a shot(s)...
    this.handle_missed_shots(stat2); // generate list of shots to fetch
    // ignore this SHOT for now; it'll be fetched later as part of the list, in-order
    return;
  }

  // OK, we are 'accepting' this shot.

  // note that we have received shot 'n'
  this.getshot_n_trycounter[shot.shotnum] = 0;
  this.shotnums_wanted[shot.shotnum] = 0;

  this.expected_shotnum = (shot.shotnum+1)%250; // next expected shot#

/*
console.log(2,'handle_shot',this.ip,'shotnum',shot.shotnum,",counters=",shot.counters
   ,"temperature_counter=",shot.temp_rdg,"status_flags=",shot.out_set,'out_set(high)=',shot.out_set&~ALL_MICS
   ,'mic1=',!!(shot.out_set & MIC1)
   ,'mic2=',!!(shot.out_set & MIC2)
   ,'mic3=',!!(shot.out_set & MIC3)
   ,'mic4=',!!(shot.out_set & MIC4)
   ,'mic5=',!!(shot.out_set & MIC5)
  );
*/
/*
The "shot.out_set" code above WORKS, e.g.:
2 'handle_shot' '192.168.0.83' 'shotnum' 8 ',counters=' [ 1291315, 1291774, 0, 1291993, 1291955 ] 'temperature_counter=' 581 'status_flags=' 36 'out_set(high)=' 32 'mic1=' false 'mic2=' false 'mic3=' true 'mic4=' false 'mic5=' false

*/



//  console.log(shot);

  // check if this shot has already been received
  var oldshot = this.oldshots[shot.shotnum];
  //if (oldshot == shot) { // mumble, grumble, can't do this...
  // .. so we have to brute-force it instead:
  if (oldshot &&
    oldshot.shotnum == shot.shotnum &&
    oldshot.counters[0] == shot.counters[0] &&
    oldshot.counters[1] == shot.counters[1] &&
    oldshot.counters[2] == shot.counters[2] &&
    oldshot.counters[3] == shot.counters[3] &&
    oldshot.counters[4] == shot.counters[4] &&
    oldshot.temp_rdg == shot.temp_rdg &&
    oldshot.out_set == shot.out_set) {
    //console.log("== hit!! #############################3");
    // this shot is a duplicate.... ignore it
//TODO:dmc
    return;
  } else {
    //console.log("== not hit!! #############################3");
  }
  this.oldshots[shot.shotnum] = shot;

  // DMC 21-Sep-2015: now tweak it a bit; replace overflowed counter values (>1,200,000) with -1
  // DMC 27-Feb-2016: use "out_set" status bits to detect overflows
  var i;
//console.log("trace handle_shot,",G1_MAX_TIMER_RDG,"timers:",
//shot.counters[0],shot.counters[1],shot.counters[2],shot.counters[3],shot.counters[4]);
  for (i = 0; i < 5; i++) if (!(shot.out_set & MIC_MASKS[i])) shot.counters[i] = -1;
//console.log("trace handle_shot AFTER,",G1_MAX_TIMER_RDG,"timers:",
//shot.counters[0],shot.counters[1],shot.counters[2],shot.counters[3],shot.counters[4]);


  this.evt_time = Date.now();
  this.elapsed_time = this.evt_time - this.send_time;
  this.evt = "shot";
  var logshot = {
    ts: this.evt_time,
    //ip: this.ip, //targetid: this.targetid,
    source: 'G1',
    shot: shot, //geom: this.geom
  }
  this.emit_shot(logshot); // G1: handle_shot

}

//var slow_poll_rate=5000, fast_poll_rate=1000;
//var slow_poll_rate=5000, fast_poll_rate=100;
//var slow_poll_rate=5000, fast_poll_rate=600;
//var slow_poll_rate=5000, fast_poll_rate=50;
//var slow_poll_rate=5000, fast_poll_rate=10;
//var slow_poll_rate=2500, fast_poll_rate=10; // DMC 17 Jan 2013
//var slow_poll_rate=2500, fast_poll_rate=400; // DMC 9 Feb
//var slow_poll_rate=2500, fast_poll_rate=10; // DMC 14 Mar 2013
//var slow_poll_rate=2500, fast_poll_rate=20; // DMC 9 Feb
//var slow_poll_rate=500, fast_poll_rate=20; // DMC 5 June 2013

//var slow_poll_rate=500, fast_poll_rate=250; // DMC 10 June 2013

//var slow_poll_rate=500, fast_poll_rate=100; // DMC 11 June 2013

// with new firmware...
var old_fw_slow_poll_rate=1100, old_fw_fast_poll_rate=70,
    new_fw_slow_poll_rate=1000, new_fw_fast_poll_rate=500; // DMC 20 June 2013

//var slow_poll_rate=2500, fast_poll_rate=800; // slower, for testing...


var dead_target_time = 3000; // if we haven't heard from the target at least once in
// this amount of time, send out a "tgt_missedpacket"

var min_heartbeat_interval = 500; // don't broadcast hearbeats quicker than this.

// TODO: add detect-missed-shots logic.


// set_interval_timer() -- sets a timer for the expected answer, typically 0.6s
TargetListener.prototype.set_interval_timer = function(poll_rate) {
  // if no change, do nothing
  if (poll_rate == this.poll_interval && this.interval_timer != undefined) {
    return;
  }
  this.clear_interval_timer();
  logdebug(4,'set_interval_timer', this.poll_interval,'==>',poll_rate,' ip=',this.ip);
  this.poll_interval = poll_rate;
  this.interval_timer = setInterval(interval_func, this.poll_interval,this);
}

TargetListener.prototype.clear_interval_timer = function() {
  if (this.interval_timer) {
    clearInterval(this.interval_timer);
    logdebug(4,'clear_interval_timer',this.ip);
  }
  this.interval_timer = undefined;
}

// tickle() is called every time we hear from the beam
TargetListener.prototype.tickle = function(msg,rinfo) {

  // once we hear from target, shift to fast polling
  var fast_poll_rate = this.new_firmware ? new_fw_fast_poll_rate : old_fw_fast_poll_rate;

  if (this.poll_interval != fast_poll_rate) {
    logdebug("IP",this.ip,"*****upping tempo from",this.poll_interval,"to",fast_poll_rate);
    logdebug("this=",this);
    this.set_interval_timer(fast_poll_rate); // speed up
  }

  this.recv_time = this.evt_time = Date.now();
//  logdebug('##tickle: set this.evt_time=',this.evt_time,', ip=',this.ip/*,'this=',this*/);
  this.elapsed_time = this.recv_time - this.send_time;


  // if we're polling really quickly, don't update status too quickly:
  if (this.last_stat_time && this.evt_time < this.last_stat_time + min_heartbeat_interval) {
//    logdebug("**tickle: min_heartbeat_interval not reached, ip=",this.ip);
//    logdebug('**tickle: min_heartbeat_interval not reached, this.last_stat_time='
//      ,this.last_stat_time,this.evt_time - this.last_stat_time);
  } else {
//    logdebug('##tickle: set this.last_stat_time=',this.last_stat_time,', ip=',this.ip);
    this.last_stat_time = this.evt_time;
    this.evt = "tickle";
    this.tgt_stat = 'tgt_there';
    this.beamevents.emit('tgt_heartbeat');
//console.log('>>>tgt_there',this.ip);
  }

  // pass this on
  this.beamevents.emit('beam_response',msg,rinfo);
}

function interval_func(tlis) {
  // evt_time     - set by tickle() and by interval_func() each time they're called
  // send_time      set by get_new_shot()-->send_cmd()
  // recv_time      set by set by tickle() whenever we receive an incoming response from beam
  // last_stat_time set every time a 'tgt_heartbeat' is emitted

  tlis.evt_time = Date.now();
  // has a timeout happened?  If we have *never* heard from beam (!recv_time),
  // or if we have more recently written to the beam than heard from it,
  // then we have a timeout.
  if (!tlis.recv_time || tlis.send_time > tlis.recv_time) {
    // timeout; we have not heard from the beam at least once since we wrote to it

    if (tlis.evt == "timeout") { // then we are not the first timeout
    } else {
      tlis.elapsed_time = 0;
    }
    tlis.elapsed_time += tlis.evt_time - tlis.send_time
    // it has been a 'good long time' since we last heard from the target.

    // if we're polling really quickly, don't update status too quickly:
    if (tlis.last_stat_time && tlis.evt_time < tlis.last_stat_time + min_heartbeat_interval) {
//      logdebug('**interval_func: min_heartbeat_interval not reached, ip=',
//        tlis.ip,'tlis.evt_time=',tlis.evt_time /*,'tlis=',tlis*/);
//    } else {
    } else if (tlis.evt_time - tlis.recv_time > dead_target_time){ // only emit 'tgt_missedpacket' if
      tlis.last_stat_time = tlis.evt_time;
      tlis.evt = "timeout";
      logdebug(5, "interval_func: timeout detected; IP=",tlis.ip);
      logdebug(9, "interval_func: timeout detected; tlis=",tlis);

      tlis.tgt_stat = 'tgt_missedpacket';
      tlis.beamevents.emit('tgt_heartbeat');
//console.log('<<<tgt_missedpacket',tlis.ip,'send_t=',tlis.send_time,'recv_t=',tlis.recv_time,'now=',Date.now());
      var slow_poll_rate = this.new_firmware ? new_fw_slow_poll_rate : old_fw_slow_poll_rate;

      if (tlis.poll_interval != slow_poll_rate) {
        logdebug("IP",tlis.ip,"*****dropping tempo from",tlis.poll_interval,"to",slow_poll_rate);
        tlis.set_interval_timer(slow_poll_rate); // slow down, if we've missed enough traffic to justify
        // telling the world about it.
      }
    }
  }
  var shotnum_to_get = tlis.shotnums_wanted.indexOf(1);
  if (shotnum_to_get != -1) {
    //if (this.getshot_n_trycounter[shotnum_to_get] > 3) { // give up??
    tlis.get_shot_n(shotnum_to_get);
    logdebug("IP",tlis.ip,'** RETRYING SHOT **, get_shot_n(',shotnum_to_get,')');
  }

  if (shotnum_to_get == -1) {
    tlis.get_new_shot();
  }
}


TargetListener.prototype.get_new_shot = function() {
  logdebug(5,"get_new_shot(",this.ip,")");
  send_cmd(SHOT_REQUEST,this,IP_PORT_UDP_SHOT_REQUEST);
}

TargetListener.prototype.get_shot_n = function(n) {
  var xtra=new Buffer(2); xtra.writeUInt16LE(n,0);
  send_cmd(SHOT_NEXT_TIME,this,IP_PORT_UDP_SHOT_REQUEST,xtra);
  // ++undefined --> NaN
  if (this.getshot_n_trycounter[n] == undefined) this.getshot_n_trycounter[n] = 0;
  this.getshot_n_trycounter[n]++;
  logdebug(3,'get_shot_n(',this.ip,',',n,'), try#=',this.getshot_n_trycounter[n]);
}

// will send event "answer_ok" if there is a response
TargetListener.prototype.ping = function() {
  send_cmd(HI_THERE,this,IP_PORT_UDP_ADMIN_REQUEST);
}

TargetListener.prototype.start_poller = function() {
  logdebug("start_poller, IP",this.ip,this.connectionType)
  this.expected_shotnum = undefined;
  this.new_firmware = undefined;

  // TODO: clear oldshots?
  if (this.connectionType === undefined || this.connectionType !== 'TCP')
    this.set_interval_timer(old_fw_slow_poll_rate); // start slowly

// DMC: perhaps G2 connection start/stop belong here?

  if ( this.connectionType === "TCP" && this.ip !== "" ) {
    this.initTcpClient(this.ip);
  }
}

g_initTcpClient_cnt = 0;

TargetListener.prototype.stop_poller = function() {
  logdebug("stop_poller, IP",this.ip,this.connectionType)
  this.clear_interval_timer();
  // TODO: perhaps announce that polling is down?

// DMC: perhaps G2 connection start/stop belong here?
  if (this.tcpClient !== undefined) {
      log.debug('stop_poller(), tcpClient.close(',this.ip,')','Client=',this.tcpClient);
      this.tcpClient.close();
      delete this.tcpClient;
  }
}
